package de.fhdw.wtf.tooling.editor;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.jface.text.TextAttribute;
import org.eclipse.jface.text.rules.IRule;
import org.eclipse.jface.text.rules.IToken;
import org.eclipse.jface.text.rules.IWordDetector;
import org.eclipse.jface.text.rules.MultiLineRule;
import org.eclipse.jface.text.rules.RuleBasedScanner;
import org.eclipse.jface.text.rules.Token;
import org.eclipse.jface.text.rules.WhitespaceRule;
import org.eclipse.jface.text.rules.WordRule;
import org.eclipse.swt.SWT;

public class WTFRuleScanner extends RuleBasedScanner {
	
	private static String[] keywords = { "class", "group", "exception" };
	private static String[] baseTypes = { "Integer", "String", "Anything" };
	private static String[] symbols = { "+", ":", "=", "*", "**", "->", ">" };
	private static String[] modifiers = { "prior", "transient", "findable", "mutable", "abstract", "service",
			"visitable", "symmetric" };
	
	/*
	 * Zeichen, die Namen und Schlüsselwörter/Basisdatentypen/Modifier begrenzen -> Erkennung dieser Tokens.
	 */
	private static Character[] nonWordCharacters = { ';', ':', '+', '=', '*', '-', '>', ' ', '[', ']', '(', ')', '{',
			'}', ',' };
	
	private final List<IRule> rules = new ArrayList<>();
	private IToken keywordToken, symbolToken, commentToken, otherToken, baseTypeToken, modifierToken;
	
	private static WTFRuleScanner instance;
	
	public static synchronized WTFRuleScanner create(final ColorManager colorManager) {
		if (instance == null) {
			instance = new WTFRuleScanner(colorManager);
		}
		return instance;
	}
	
	private WTFRuleScanner(final ColorManager colorManager) {
		this.createTokens(colorManager);
		this.generateStandardRules();
	}
	
	/*
	 * Tokens legen die Formatierung in Eclipse fest.
	 */
	private void createTokens(final ColorManager colorManager) {
		this.keywordToken =
				new Token(new TextAttribute(colorManager.getColor(WTFColorConstants.WTF_KEYWORD),
						colorManager.getColor(WTFColorConstants.BACKGROUND), SWT.BOLD));
		this.symbolToken =
				new Token(new TextAttribute(colorManager.getColor(WTFColorConstants.WTF_SYMBOLS),
						colorManager.getColor(WTFColorConstants.BACKGROUND), SWT.BOLD));
		this.commentToken = new Token(new TextAttribute(colorManager.getColor(WTFColorConstants.WTF_COMMENT)));
		this.otherToken = new Token(new TextAttribute(colorManager.getColor(WTFColorConstants.OTHER)));
		this.baseTypeToken =
				new Token(new TextAttribute(colorManager.getColor(WTFColorConstants.BASETYPE),
						colorManager.getColor(WTFColorConstants.BACKGROUND), SWT.BOLD));
		this.modifierToken =
				new Token(new TextAttribute(colorManager.getColor(WTFColorConstants.MODIFIERS),
						colorManager.getColor(WTFColorConstants.BACKGROUND), SWT.ITALIC));
	}
	
	/*
	 * Erzeugt Regeln für Eclipse-Syntaxerkennung.
	 */
	public void generateStandardRules() {
		this.rules.clear();
		final IRule[] delRule = new IRule[this.rules.size()];
		this.setRules(delRule);
		
		this.rules.add(new MultiLineRule("#", "#", this.commentToken));
		this.rules.add(new WhitespaceRule(new WTFWhitespaceDetector()));
		
		final WordRule keywordRule = this.createRule(keywords, this.keywordToken);
		
		final WordRule symbolRule = this.createRule(symbols, this.symbolToken);
		
		final WordRule baseTypeRule = this.createRule(baseTypes, this.baseTypeToken);
		
		final WordRule modifierRule = this.createRule(modifiers, this.modifierToken);
		
		this.rules.add(modifierRule);
		this.rules.add(baseTypeRule);
		this.rules.add(keywordRule);
		this.rules.add(symbolRule);
		final IRule[] result = new IRule[this.rules.size()];
		this.rules.toArray(result);
		this.setRules(result);
	}
	
	/*
	 * Erzeugt Regel für Worterkennung.
	 */
	private WordRule createRule(final String[] words, final IToken token) {
		
		final WordRule rule = new WordRule(new IWordDetector() {
			
			@Override
			public boolean isWordStart(final char c) {
				for (final String word : words) {
					if (word.length() > 0 && word.toCharArray()[0] == c) {
						return true;
					}
				}
				return false;
			}
			
			@Override
			public boolean isWordPart(final char c) {
				
				for (final char character : nonWordCharacters) {
					if (c == character) {
						return false;
					}
				}
				
				return true;
			}
		}, this.otherToken);
		
		for (final String word : words) {
			rule.addWord(word, token);
		}
		
		return rule;
	}
	
	/*
	 * Hilfsmethode zum Setzen der Regeln.
	 */
	public void updateRules(final WordRule rule) {
		this.rules.add(rule);
		final IRule[] result = new IRule[this.rules.size()];
		this.rules.toArray(result);
		this.setRules(result);
	}
}
